import os
from .helpers import (ns_os_environ_get, parse_boolean, int_or_none)

__all__ = [
    'LOG_LEVEL', 'TAPISPY_LOG_LEVEL', 'SWAGGERPY_LOG_LEVEL', 'SHOW_CURL',
    'VERBOSE_ERRORS'
]

DEFAULT_LOG_LEVEL = 'WARNING'
LOG_LEVEL = ns_os_environ_get('LOG_LEVEL', DEFAULT_LOG_LEVEL)
TAPISPY_LOG_LEVEL = ns_os_environ_get('TAPISPY_LOG_LEVEL', LOG_LEVEL)
SWAGGERPY_LOG_LEVEL = ns_os_environ_get('SWAGGERPY_LOG_LEVEL',
                                        TAPISPY_LOG_LEVEL)
SHOW_CURL = parse_boolean(ns_os_environ_get('SHOW_CURL', '0'))
VERBOSE_ERRORS = parse_boolean(ns_os_environ_get('VERBOSE_ERRORS', '1'))



from .agave import Agave

# from . import sessions
from . import tenants

import sys  # noqa
from agavepy import asynchronous  # noqa
sys.modules['agavepy.async'] = asynchronous  # noqa
from agavepy import interactive  # noqa
from agavepy import settings  # noqa



__all__ = ['TokenCommands']


class TokenCommands(object):
    def refresh(self):
        """If possible, attempt to refresh the Oauth token

        This is the function that should be used to 
        regenerate an Oauth token as it can deal with 
        cases where no refresh capability is configured. 
        """
        if getattr(self, 'token') is not None:
            return self.token.refresh()
        else:
            if not self.can_refresh:
                return getattr(self, '_token', None)
            else:
                raise Exception(
                    'Oauth client is not configured to refresh tokens')

    def tokens(self):
        """Return current access and refresh tokens
        """
        resp = {}
        resp['access_token'] = getattr(self, '_token', None)
        resp['refresh_token'] = getattr(self, 'refresh_token', None)
        return resp



import os

ENV_PREFIX = 'TAPIS_PY'

BOOLEAN_TRUE_STRINGS = ('true', 'on', 'ok', 'y', 'yes', '1')
BOOLEAN_FALSE_STRINGS = ('false', 'off', 'n', 'no', '0')

__all__ = [
    'ns_os_environ_get', 'fix_assets_path', 'array_from_string',
    'set_from_string', 'parse_boolean', 'int_or_none', 'ENV_PREFIX'
]


def ns_os_environ_get(env_var_name, default, prefix=ENV_PREFIX):
    if prefix is not None:
        if not prefix.endswith('_'):
            prefix = prefix + '_'
    varname = '{}{}'.format(prefix, env_var_name)
    resp = os.environ.get(varname, default)
    return resp


def fix_assets_path(path):
    fullpath = os.path.join(os.path.dirname(__file__), "../", path)
    return fullpath


def array_from_string(s):
    array = s.split(',')
    if "" in array:
        array.remove("")
    return array


def set_from_string(s):
    return set(array_from_string(s))


def parse_boolean(s):
    """Takes a string and returns the equivalent as a boolean value."""
    s = s.strip().lower()
    if s in BOOLEAN_TRUE_STRINGS:
        return True
    elif s in BOOLEAN_FALSE_STRINGS:
        return False
    else:
        raise ValueError('Invalid boolean value %r' % s)


def int_or_none(value):
    if value is None:
        return value
    return int(value)



#
# Copyright (c) 2013, Digium, Inc.
#
"""Code for handling the base Swagger API model.
"""

import json
import os
import urllib.request, urllib.parse, urllib.error
import urllib.parse

from swaggerpy.http_client import SynchronousHttpClient
from swaggerpy.processors import SwaggerProcessor, SwaggerError

SWAGGER_VERSIONS = ["1.1", "1.2"]

SWAGGER_PRIMITIVES = [
    'void',
    'string',
    'boolean',
    'number',
    'int',
    'long',
    'double',
    'float',
    'Date',
]


# noinspection PyDocstring
class ValidationProcessor(SwaggerProcessor):
    """A processor that validates the Swagger model.
    """
    def process_resource_listing(self, resources, context):
        required_fields = ['apis', 'swaggerVersion']
        validate_required_fields(resources, required_fields, context)

        if not resources['swaggerVersion'] in SWAGGER_VERSIONS:
            raise SwaggerError(
                "Unsupported Swagger version %s" % resources.swaggerVersion,
                context)

    def process_resource_listing_api(self, resources, listing_api, context):
        validate_required_fields(listing_api, ['path', 'description'], context)

        if not listing_api['path'].startswith("/"):
            raise SwaggerError("Path must start with /", context)

    def process_api_declaration(self, resources, resource, context):
        required_fields = ['swaggerVersion', 'resourcePath', 'apis', 'models']
        validate_required_fields(resource, required_fields, context)
        # Check model name and id consistency
        for (model_name, model) in list(resource['models'].items()):
            if model_name != model['id']:
                raise SwaggerError("Model id doesn't match name", context)
                # Convert models dict to list

    def process_resource_api(self, resources, resource, api, context):
        required_fields = ['path', 'operations']
        validate_required_fields(api, required_fields, context)

    def process_operation(self, resources, resource, api, operation, context):
        required_fields = ['method', 'nickname']
        validate_required_fields(operation, required_fields, context)

    def process_parameter(self, resources, resource, api, operation, parameter,
                          context):
        required_fields = ['name', 'paramType']
        validate_required_fields(parameter, required_fields, context)
        if parameter['paramType'] == 'path':
            # special handling for path parameters
            parameter['required'] = True
            parameter['type'] = 'string'
        else:
            # dataType is required for non-path parameters
            validate_required_fields(parameter, ['type'], context)
        if 'allowedValues' in parameter:
            raise SwaggerError(
                "Field 'allowedValues' invalid; use 'allowableValues'",
                context)

    def process_error_response(self, resources, resource, api, operation,
                               error_response, context):
        required_fields = ['code', 'reason']
        validate_required_fields(error_response, required_fields, context)

    def process_model(self, resources, resource, model, context):
        required_fields = ['id', 'properties']
        validate_required_fields(model, required_fields, context)
        # Move property field name into the object
        for (prop_name, prop) in list(model['properties'].items()):
            prop['name'] = prop_name

    def process_property(self, resources, resource, model, prop, context):
        required_fields = ['type']
        validate_required_fields(prop, required_fields, context)


def json_load_url(http_client, url):
    """Download and parse JSON from a URL.

    :param http_client: HTTP client interface.
    :type  http_client: http_client.HttpClient
    :param url: URL for JSON to parse
    :return: Parsed JSON dict
    """
    scheme = urllib.parse.urlparse(url).scheme
    if scheme == 'file':
        # requests can't handle file: URLs
        fp = urllib.request.urlopen(url)
        try:
            return json.load(fp)
        finally:
            fp.close()
    else:
        resp = http_client.request('GET', url)
        resp.raise_for_status()
        return resp.json()


class Loader(object):
    """Abstraction for loading Swagger API's.

    :param http_client: HTTP client interface.
    :type  http_client: http_client.HttpClient
    :param processors: List of processors to apply to the API.
    :type  processors: list of SwaggerProcessor
    """
    def __init__(self, http_client, processors=None):
        self.http_client = http_client
        if processors is None:
            processors = []
            # always go through the validation processor first
        # noinspection PyTypeChecker
        self.processors = [ValidationProcessor()] + processors

    def load_resource_listing(self, resources_url, base_url=None):
        """Load a resource listing, loading referenced API declarations.

        The following fields are added to the resource listing object model.
         * ['url'] = URL resource listing was loaded from
         * The ['apis'] array is modified according to load_api_declaration()

        The Loader's processors are applied to the fully loaded resource
        listing.

        :param resources_url:   File name for resources.json
        :param base_url:    Optional URL to be the base URL for finding API
                            declarations. If not specified, 'basePath' from the
                            resource listing is used.
        """

        # Load the resource listing
        resource_listing = json_load_url(self.http_client, resources_url)

        # Some extra data only known about at load time
        resource_listing['url'] = resources_url
        if not base_url:
            base_url = resource_listing.get('basePath', resources_url)

        # Load the API declarations
        for api in resource_listing.get('apis'):
            self.load_api_declaration(base_url, api)

        # Now that the raw object model has been loaded, apply the processors
        self.process_resource_listing(resource_listing)
        return resource_listing

    def load_api_declaration(self, base_url, api_dict):
        """Load an API declaration file.

        api_dict is modified with the results of the load:
         * ['url'] = URL api declaration was loaded from
         * ['api_declaration'] = Parsed results of the load

        :param base_url: Base URL to load from
        :param api_dict: api object from resource listing.
        """
        path = api_dict.get('path').replace('{format}', 'json')
        api_dict['url'] = urllib.parse.urljoin(base_url + '/', path.strip('/'))
        api_dict['api_declaration'] = json_load_url(self.http_client,
                                                    api_dict['url'])

    def process_resource_listing(self, resources):
        """Apply processors to a resource listing.

        :param resources: Resource listing to process.
        """
        for processor in self.processors:
            processor.apply(resources)


def validate_required_fields(json, required_fields, context):
    """Checks a JSON object for a set of required fields.

    If any required field is missing, a SwaggerError is raised.

    :param json: JSON object to check.
    :param required_fields: List of required fields.
    :param context: Current context in the API.
    """
    missing_fields = [f for f in required_fields if not f in json]

    if missing_fields:
        raise SwaggerError("Missing fields: %s" % ', '.join(missing_fields),
                           context)


def load_file(resource_listing_file, http_client=None, processors=None):
    """Loads a resource listing file, applying the given processors.

    :param http_client: HTTP client interface.
    :param resource_listing_file: File name for a resource listing.
    :param processors:  List of SwaggerProcessors to apply to the resulting
                        resource.
    :return: Processed object model from
    :raise: IOError: On error reading api-docs.
    """
    file_path = os.path.abspath(resource_listing_file)
    url = urllib.parse.urljoin('file:', urllib.request.pathname2url(file_path))
    # When loading from files, everything is relative to the resource listing
    dir_path = os.path.dirname(file_path)
    base_url = urllib.parse.urljoin('file:',
                                    urllib.request.pathname2url(dir_path))
    return load_url(url,
                    http_client=http_client,
                    processors=processors,
                    base_url=base_url)


def load_url(resource_listing_url,
             http_client=None,
             processors=None,
             base_url=None):
    """Loads a resource listing, applying the given processors.

    :param resource_listing_url: URL for a resource listing.
    :param http_client: HTTP client interface.
    :param processors:  List of SwaggerProcessors to apply to the resulting
                        resource.
    :param base_url:    Optional URL to be the base URL for finding API
                        declarations. If not specified, 'basePath' from the
                        resource listing is used.
    :return: Processed object model from
    :raise: IOError, URLError: On error reading api-docs.
    """
    if http_client is None:
        http_client = SynchronousHttpClient()

    loader = Loader(http_client=http_client, processors=processors)
    return loader.load_resource_listing(resource_listing_url,
                                        base_url=base_url)


def load_json(resource_listing, http_client=None, processors=None):
    """Process a resource listing that has already been parsed.

    :param resource_listing: Parsed resource listing.
    :type  resource_listing: dict
    :param http_client:
    :param processors:
    :return: Processed resource listing.
    """
    if http_client is None:
        http_client = SynchronousHttpClient()

    loader = Loader(http_client=http_client, processors=processors)
    loader.process_resource_listing(resource_listing)
    return resource_listing



from __future__ import absolute_import
import dateutil.parser
import logging
import os
import requests
import sys

from agavepy import settings
from agavepy.aloe import EXCEPTION_MODELS
from agavepy.errors import AgaveException, AgaveError
from agavepy.util import AttrDict, with_refresh

sys.path.insert(0, os.path.dirname(__file__))  # noqa
from .swaggerpy.processors import SwaggerProcessor  # noqa
from .swaggerpy.http_client import SynchronousHttpClient  # noqa
from .swaggerpy.client import SwaggerClient  # noqa

HERE = os.path.dirname(os.path.abspath(__file__))
logger = logging.getLogger(__name__)
logging.getLogger(__name__).setLevel(
    os.environ.get('TAPISPY_LOG_LEVEL', logging.WARNING))

__all__ = [
    'AgaveProcessor', 'Operation', 'Resource', 'SwaggerProcessor',
    'SynchronousHttpClient', 'SwaggerClient'
]


class Resource(object):
    def __init__(self, resource, client):
        """

        :param resource:
        :param client: the Agave instance associated with this resource
        :return:
        """
        self.resource = resource
        self.client = client

    def __getattr__(self, attr):
        return Operation(self.resource, attr, client=self.client)

    def __dir__(self):
        if hasattr(self, "clients_resource") and hasattr(
                self.clients_resource, "resources"):
            clients = self.client.clients_resource.resources
            base = list(clients[self.resource].operations.keys())
        else:
            base = []
        if self.client.all is not None:
            base.extend(
                list(self.client.all.resources[
                    self.resource].operations.keys()))
        return base


class Operation(object):

    PRIMITIVE_TYPES = ["array", "string", "integer", "int", "boolean", "dict"]

    def __init__(self, resource, operation, client):
        self.resource = resource
        self.operation = operation
        self.client = client
        self.models = self.get_models()
        self.return_type = self.get_return_type()

    def get_base(self):
        """Get the base ari for this resource."""

        return (self.client.clients_resource
                if self.resource == "clients" else self.client.all)

    def get_operation(self):
        base = self.get_base()
        return getattr(getattr(base, self.resource), self.operation)

    def get_models(self):
        """Get JSON with all the models declarations."""

        base = self.get_base()
        return getattr(base, self.resource).json["api_declaration"]["models"]

    def get_return_type(self):
        """Get JSON with the return type for this operation."""

        return self.get_operation().json

    def __call__(self, *args, **kwargs):
        def operation():
            f = self.get_operation()
            response = f(*args, **kwargs)
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as h:
                if settings.VERBOSE_ERRORS:
                    # Extract the API JSON message and attach it
                    # to the HTTPError object before raising it
                    code = h.response.status_code
                    reason = h.response.reason + ' for ' + h.response.url
                    try:
                        message = h.response.json().get('message')
                    except Exception:
                        message = h.response.text
                    raise requests.exceptions.HTTPError(code,
                                                        reason,
                                                        message,
                                                        response=h.response,
                                                        request=h.request)
                else:
                    raise h
            return response

        kwargs["proxies"] = self.client.proxies
        logger.debug('Operation.__call__()...')
        resp = with_refresh(self.client, operation)
        if resp.ok:
            # if response is 204 (no content) return none directly
            if resp.status_code == 204 and not resp.content:
                return resp
            # if response is raw file, return it directly
            if self.resource == "files" and (
                    self.operation == "download"
                    or self.operation == "downloadFromDefaultSystem"):
                return resp
            # if response is a result, return it directly as well:
            if self.resource == "actors" and self.operation == "getOneExecutionResult":
                return resp
            # get the version associated with this response; the version is used for post_processing, as
            # some response models depend on the version.
            try:
                version = resp.json().get("version")
            except Exception:
                version = None
            processed = self.post_process(resp.json(), self.return_type,
                                          version)
            result = processed["result"] if "result" in processed else None
            # if operation is clients.create, save name
            if self.resource == "clients" and self.operation == "create":
                self.client.set_client(result["consumerKey"],
                                       result["consumerSecret"])
            return result
        else:
            # if the response is not 2xx return the unprocessed json rather
            # than raising an exception, since the return json contains
            # the error message
            raise AgaveException(resp.json())

    def post_process(self, obj, return_type, version):
        if return_type is None:
            return self.process_untyped(obj)

        type_name = return_type["type"].lower()
        if type_name in self.PRIMITIVE_TYPES:
            f = getattr(self, "process_{}".format(type_name))
            try:
                return f(obj, return_type, version)
            except Exception:
                return self.process_untyped(obj)
        return self.process_model(obj, return_type, version)

    def process_untyped(self, obj, version=None):
        return obj

    def process_dict(self, obj, return_type, version=None):
        return obj

    def process_array(self, obj, return_type, version):
        items = return_type["items"]
        items_type = items.get("type", items.get("$ref"))
        return [
            self.post_process(elem, {"type": items_type}, version)
            for elem in obj
        ]

    def process_string(self, obj, return_type, version=None):
        if obj is None:
            # why is agave returning null for a string type?
            return obj
        if return_type.get("format") == "date-time":
            return dateutil.parser.parse(obj)
        return obj

    def process_integer(self, obj, return_type, version=None):
        return obj

    process_int = process_integer

    def process_boolean(self, obj, return_type, version=None):
        return obj

    def get_model_spec(self, model_name, version):
        """
        Look up the model_spec associated with a model name and check for exceptions based on the version of
         the response.
        :param model_name:
        :return:
        """
        if model_name in EXCEPTION_MODELS:
            version_cutoff = EXCEPTION_MODELS[model_name]["version_cutoff"]
            if version > version_cutoff:
                # the API version of this response was greater than the version cutoff, so we need to look up the
                # the model spec in the exception_resources
                new_model_name = EXCEPTION_MODELS[model_name]["model"]
                return self.client.resource_exceptions[new_model_name][
                    "properties"]
        # if model was not an exception model OR the version was less than the version cutoff, just return the
        # the model in the original models object -
        return self.models[model_name]["properties"]

    def process_model(self, obj, return_type, version):
        model_name = return_type["type"]

        model_spec = self.get_model_spec(model_name, version)
        result = AttrDict({})
        for k in obj:
            try:
                result[k] = self.post_process(obj[k], model_spec.get(k),
                                              version)
            except Exception:
                result[k] = obj
        return result


class AgaveProcessor(SwaggerProcessor):
    def process_property(self, resources, resource, model, prop, context):
        if prop.get("format", None) == "date-time":
            pass

    def process_model(self, resources, resource, model, context):
        pass



import time


class Error(Exception):
    pass


class TimeoutError(Error):
    pass


class AgaveAsyncResponse(object):
    """
    Implements parts of the concurrent.futures.Future interface for nonblocking Agave responses.
    """
    def __init__(self, ag, response):
        """Construct an asynchronous response object from an Agave response which should be an agavepy.agave.AttrDict
        """
        self.ag = ag
        self.response = response
        self.status = response.status
        self.retries = 0
        self.url = response._links.get('history').get('href')
        if not self.url:
            raise Error(
                "Error parsing response object: no URL detected. response: " +
                str(response))
        # url's returned by agave sometimes point to docker.example.com
        self.url = self.url.replace('https://docker.example.com',
                                    self.ag.api_server)
        # url's returned by agave sometimes have the version as 2.0, so we replace that here
        self.url = self.url.replace('/2.0/', '/v2/')

    def _update_status(self):
        rsp = self.ag.geturl(self.url)
        if (rsp.status_code == 404
                or rsp.status_code == 403) and self.retries < 10:
            time.sleep(1.5)
            self.retries += 1
            return self._update_status()
        if not rsp.status_code == 200:
            raise Error("Error updating status; invalid status code:  " +
                        str(rsp.status_code) + str(rsp.content))
        result = rsp.json().get('result')
        if not result:
            raise Error("Error updating status: " + str(rsp))
        # if the transfer ever completed, we'll call it good:
        if len([x for x in result if 'COMPLETE' in x['status']]) > 0:
            self.status = 'FINISHED'
        # jobs end in 'finished' status
        elif len([x for x in result if 'FINISHED' in x['status']]) > 0:
            self.status = 'FINISHED'
        # job ended in 'failed' status
        elif len([x for x in result if 'FAILED' in x['status']]) > 0:
            self.status = 'FAILED'
        else:
            # sort on creation time of the history object
            result = sorted(result, key=lambda k: k['created'])
            self.status = result[0].get('status')

    def _is_done(self):
        return self.status == 'FINISHED' or self.status == 'FAILED'

    def done(self):
        """Return True if the call was successfully cancelled or finished running."""
        self._update_status()
        return self._is_done()

    def result(self, timeout=None):
        """
        Returns the result of the original call, blocking until the result is returned or the timeout parameter is
        reached. The timeout paramter is interpreted in seconds.
        :param timeout:
        :return:
        """
        if self._is_done():
            return self.status
        self._update_status()
        now = time.time()
        if timeout:
            future = now + timeout
        else:
            future = float("inf")
        while time.time() < future:
            time.sleep(1)
            self._update_status()
            if self._is_done():
                return self.status
        raise TimeoutError()



#
# Copyright (c) 2013, Digium, Inc.
#
"""Swagger client library.
"""
from __future__ import print_function
import curlify
import json
import logging
import os.path
import re
import urllib.request
import urllib.parse
import urllib.error
import swaggerpy
import sys
from numbers import Real

from requests_toolbelt import MultipartEncoder

from swaggerpy.http_client import SynchronousHttpClient
from swaggerpy.processors import WebsocketProcessor, SwaggerProcessor

sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))  # noqa

LOG_LEVEL = os.environ.get('SWAGGERPY_LOG_LEVEL',
                           os.environ.get('TAPISPY_LOG_LEVEL', logging.ERROR))
logging.getLogger(__name__).setLevel(LOG_LEVEL)
logging.getLogger("urllib3").setLevel(LOG_LEVEL)
log = logging.getLogger(__name__)


def print_stderr(message):
    # TODO - Support Python2 implementation
    print('{0}'.format(message), file=sys.stderr)


class ClientProcessor(SwaggerProcessor):
    """Enriches swagger models for client processing.
    """
    def process_resource_listing_api(self, resources, listing_api, context):
        """Add name to listing_api.

        :param resources: Resource listing object
        :param listing_api: ResourceApi object.
        :type context: ParsingContext
        :param context: Current context in the API.
        """
        name, ext = os.path.splitext(os.path.basename(listing_api['path']))
        listing_api['name'] = name


class Operation(object):
    """Operation object.
    """
    def __init__(self, uri, operation, http_client, show_curl=False):
        self.uri = uri
        self.json = operation
        self.http_client = http_client
        self.show_curl = show_curl

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.json['nickname'])

    def __show_curl(self, response_object):
        """Render an HTTP request as a cURL to STDERR
        """
        if self.show_curl:
            try:
                curl_text = curlify.to_curl(response_object.request)
            except Exception as err:
                curl_text = 'Failed to render curl command: {0}'.format(err)
            print_stderr(curl_text)

    def file_like(self, obj):
        """Try to decide if we should put this object in multipart."""

        primitives = (str, Real)
        return not isinstance(obj, primitives)

    def __call__(self, **kwargs):
        """Invoke ARI operation.

        :param kwargs: ARI operation arguments.
        :return: Implementation specific response or WebSocket connection
        """
        log.info("%s?%r" %
                 (self.json['nickname'], urllib.parse.urlencode(kwargs)))
        # http method must be native string (for Python 2.x, bytes). This must be done to prevent errors when trying to
        # upload binary data because when unicode is passed, the concatenated request string gets decoded using the
        # default encoding (ASCII) and that breaks on binary data. See the bottom of this thread:
        # https://github.com/kennethreitz/requests/issues/1252
        method = str(self.json['method'])
        uri = self.uri
        params = {}
        data = {}
        headers = {}
        files = {}
        proxies = kwargs.pop('proxies')
        # allow passing custom headers
        if kwargs.get('headers'):
            try:
                headers.update(kwargs.pop('headers'))
            except ValueError:
                raise AssertionError("Parameter headers must be of type dict.")
        # allow passing a custom query dict
        if kwargs.get('query'):
            try:
                params.update(kwargs.pop('query'))
            except ValueError:
                raise AssertionError("Parameter query must be of type dict.")
        accepts_multipart = ('multipart/form-data' in self.json.get(
            'consumes', []))
        # allow passing an `x-nonce`
        if kwargs.get('nonce'):
            nonce = kwargs.get('nonce')
            params['x-nonce'] = nonce
            del kwargs['nonce']

        for param in self.json.get('parameters', []):
            pname = param['name']
            ptype = param['type']
            value = kwargs.get(pname)
            if ptype == 'dict':
                if value and param['paramType'] == 'query':
                    try:
                        params.update(value)
                    except ValueError:
                        raise AssertionError(
                            "Parameter {} must be of type dict.".format(pname))
            # Turn list params into comma separated values
            if isinstance(value, list):
                value = ",".join(value)
            if value is not None:
                param_type = param['paramType']
                if param_type == 'path':
                    uri = uri.replace('{%s}' % pname, str(value))
                elif param_type == 'query':
                    params[pname] = value
                elif param_type == 'form':
                    if accepts_multipart and self.file_like(value):
                        # AD-1345 : issue uploading large files (see ticket).
                        try:
                            file_name = os.path.basename(value.name)
                        except AttributeError:
                            raise TypeError(
                                "File upload object must have a name attribute."
                            )
                        m = MultipartEncoder(
                            fields={pname: (file_name, value, 'text/plain')})
                        headers['Content-type'] = m.content_type
                        data = m
                    else:
                        data[pname] = value
                elif param_type == 'body':
                    isjson = True
                    if isinstance(value, str):
                        data = value
                    else:
                        try:
                            data = json.dumps(value)
                        except TypeError:
                            data = value
                            if not headers.get('Content-type') \
                                    or headers.get('Content-type') == 'application/octet-stream':
                                isjson = False
                                headers[
                                    'Content-type'] = 'application/octet-stream'
                    if isjson:
                        headers['Content-type'] = 'application/json'
                else:
                    raise AssertionError("Unsupported paramType %s" %
                                         param_type)
                del kwargs[pname]
            else:
                if param['required']:
                    raise TypeError(
                        "Missing required parameter '%s' for '%s'" %
                        (pname, self.json['nickname']))
        if method.lower() == 'get':
            # look for the search dictionary on GET requests:
            value = kwargs.get('search')
            if value:
                if not isinstance(value, dict):
                    raise TypeError("search parameter must be of type dict")
                for k, v in list(value.items()):
                    params[k] = v
                kwargs.pop('search')
            value = kwargs.get('filter')
            if value:
                if not isinstance(value, str):
                    raise TypeError("filter parameter must be of type str")
                params['filter'] = value
                kwargs.pop('filter')

        if kwargs:
            raise TypeError("'%s' does not have parameters %r" %
                            (self.json['nickname'], list(kwargs.keys())))

        log.info("%s %s(%r)", method, uri, params)
        if self.json['is_websocket']:
            # Fix up http: URLs
            uri = re.sub('^http', "ws", uri)
            return self.http_client.ws_connect(uri, params=params)
        else:
            req = self.http_client.request(method,
                                           uri,
                                           params=params,
                                           data=data,
                                           headers=headers,
                                           files=files,
                                           proxies=proxies)
            self.__show_curl(req)
            return req


class Resource(object):
    """Swagger resource, described in an API declaration.

    :param resource: Resource model
    :param http_client: HTTP client API
    """
    def __init__(self, resource, http_client, show_curl=False):
        log.debug("Building resource '%s'" % resource['name'])
        self.json = resource
        decl = resource['api_declaration']
        self.http_client = http_client
        self.operations = {
            oper['nickname']: self._build_operation(decl,
                                                    api,
                                                    oper,
                                                    show_curl=show_curl)
            for api in decl['apis'] for oper in api['operations']
        }

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.json['name'])

    def __getattr__(self, item):
        """Promote operations to be object fields.

        :param item: Name of the attribute to get.
        :rtype: Resource
        :return: Resource object.
        """
        op = self.get_operation(item)
        if not op:
            raise AttributeError("Resource '%s' has no operation '%s'" %
                                 (self.get_name(), item))
        return op

    def get_operation(self, name):
        """Gets the operation with the given nickname.

        :param name: Nickname of the operation.
        :rtype:  Operation
        :return: Operation, or None if not found.
        """
        return self.operations.get(name)

    def get_name(self):
        """Returns the name of this resource.

        Name is derived from the filename of the API declaration.

        :return: Resource name.
        """
        return self.json.get('name')

    def _build_operation(self, decl, api, operation, show_curl=False):
        """Build an operation object

        :param decl: API declaration.
        :param api: API entry.
        :param operation: Operation.
        """
        log.debug("Building operation %s.%s" %
                  (self.get_name(), operation['nickname']))
        uri = decl['basePath'] + api['path']
        return Operation(uri, operation, self.http_client, show_curl=show_curl)


class SwaggerClient(object):
    """Client object for accessing a Swagger-documented RESTful service.

    :param url_or_resource: Either the parsed resource listing+API decls, or
                            its URL.
    :type url_or_resource: dict or str
    :param http_client: HTTP client API
    :type  http_client: HttpClient
    """
    def __init__(self,
                 url_or_resource,
                 http_client=None,
                 extra_processors=None,
                 show_curl=False):
        if not http_client:
            http_client = SynchronousHttpClient()
        self.http_client = http_client

        processors = [WebsocketProcessor(), ClientProcessor()]
        if extra_processors is not None:
            processors.extend(extra_processors)
        loader = swaggerpy.Loader(http_client, processors)

        if isinstance(url_or_resource, str):
            log.debug("Loading from %s" % url_or_resource)
            self.api_docs = loader.load_resource_listing(url_or_resource)
        else:
            log.debug("Loading from %s" % url_or_resource.get('basePath'))
            self.api_docs = url_or_resource
            loader.process_resource_listing(self.api_docs)

        self.resources = {
            resource['name']: Resource(resource,
                                       http_client,
                                       show_curl=show_curl)
            for resource in self.api_docs['apis']
        }

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.api_docs['basePath'])

    def __getattr__(self, item):
        """Promote resource objects to be client fields.

        :param item: Name of the attribute to get.
        :return: Resource object.
        """
        resource = self.get_resource(item)
        if not resource:
            raise AttributeError("API has no resource '%s'" % item)
        return resource

    def close(self):
        """Close the SwaggerClient, and underlying resources.
        """
        self.http_client.close()

    def get_resource(self, name):
        """Gets a Swagger resource by name.

        :param name: Name of the resource to get
        :rtype: Resource
        :return: Resource, or None if not found.
        """
        return self.resources.get(name)



#!/usr/bin/env python

#
# Copyright (c) 2013, Digium, Inc.
#
"""HTTP client abstractions.
"""

import logging
import requests
import requests.auth
import urllib.parse
import websocket

log = logging.getLogger(__name__)


class HttpClient(object):
    """Interface for a minimal HTTP client.
    """
    def close(self):
        """Close this client resource.
        """
        raise NotImplementedError("%s: Method not implemented",
                                  self.__class__.__name__)

    def request(self, method, url, params=None, data=None):
        """Issue an HTTP request.

        :param method: HTTP method (GET, POST, DELETE, etc.)
        :type  method: str
        :param url: URL to request
        :type  url: str
        :param params: Query parameters (?key=value)
        :type  params: dict
        :param data: Request body
        :type  data: Dictionary, bytes, or file-like object
        :return: Implementation specific response object
        """
        raise NotImplementedError("%s: Method not implemented",
                                  self.__class__.__name__)

    def ws_connect(self, url, params=None):
        """Create a WebSocket connection.

        :param url: WebSocket URL.
        :type  url: str
        :param params: Query parameters (?key=value)
        :type  params: dict
        :return: Implmentation specific WebSocket connection object
        """
        raise NotImplementedError("%s: Method not implemented",
                                  self.__class__.__name__)

    def set_basic_auth(self, host, username, password):
        """Configures client to use HTTP Basic authentication.

        :param host: Hostname to limit authentication to.
        :param username: Username
        :param password: Password
        """
        raise NotImplementedError("%s: Method not implemented",
                                  self.__class__.__name__)

    def set_api_key(self, host, api_key, param_name='api_key'):
        """Configures client to use api_key authentication.

        The api_key is added to every query parameter sent.

        :param host: Hostname to limit authentication to.
        :param api_key: Value for api_key.
        :param param_name: Parameter name to use in query string.
        """
        raise NotImplementedError("%s: Method not implemented",
                                  self.__class__.__name__)


class Authenticator(object):
    """Authenticates requests.

    :param host: Host to authenticate for.
    """
    def __init__(self, host):
        self.host = host

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.host)

    def matches(self, url):
        """Returns true if this authenticator applies to the given url.

        :param url: URL to check.
        :return: True if matches host, port and scheme, False otherwise.
        """
        split = urllib.parse.urlsplit(url)
        return self.host == split.hostname

    def apply(self, request):
        """Apply authentication to a request.

        :param request: Request to add authentication information to.
        """
        raise NotImplementedError("%s: Method not implemented",
                                  self.__class__.__name__)


# noinspection PyDocstring
class BasicAuthenticator(Authenticator):
    """HTTP Basic authenticator.

    :param host: Host to authenticate for.
    :param username: Username.
    :param password: Password
    """
    def __init__(self, host, username, password):
        super(BasicAuthenticator, self).__init__(host)
        self.auth = requests.auth.HTTPBasicAuth(username, password)

    def apply(self, request):
        request.auth = self.auth


# noinspection PyDocstring
class ApiKeyAuthenticator(Authenticator):
    """?api_key authenticator.

    This authenticator adds a query parameter to specify an API key.

    :param host: Host to authenticate for.
    :param api_key: API key.
    :param param_name: Query parameter specifying the API key.
    """
    def __init__(self, host, api_key, param_name='api_key'):
        super(ApiKeyAuthenticator, self).__init__(host)
        self.param_name = param_name
        self.api_key = api_key

    def apply(self, request):
        request.params[self.param_name] = self.api_key


class TokenAuthenticator(Authenticator):
    def __init__(self, host, token):
        super(TokenAuthenticator, self).__init__(host)
        self.token = token

    def apply(self, request):
        request.headers['Authorization'] = 'Bearer {}'.format(self.token)


class NonceAuthenticator(Authenticator):
    """ The NonceAuthenticator can be used when no authentication credentials are present at the time
    of the client instantiation. The assumption is that a nonce or token will be provided at request
    time to authenticate.
    """
    def __init__(self, host):
        super(NonceAuthenticator, self).__init__(host)

    def apply(self, request):
        pass


class JwtAuthenticator(Authenticator):
    def __init__(self, host, header_name, jwt):
        super(JwtAuthenticator, self).__init__(host)
        self.jwt = jwt
        self.header_name = header_name

    def apply(self, request):
        request.headers[self.header_name] = self.jwt
        request.url = request.url.strip('/v2')


# noinspection PyDocstring
class SynchronousHttpClient(HttpClient):
    """Synchronous HTTP client implementation.
    """
    def __init__(self, verify=None):
        self.session = requests.Session()
        self.authenticator = None
        self.websockets = set()
        self.verify = verify

    def close(self):
        self.session.close()
        # There's no WebSocket factory to close; close connections individually

    def set_basic_auth(self, host, username, password):
        self.authenticator = BasicAuthenticator(host=host,
                                                username=username,
                                                password=password)

    def set_api_key(self, host, api_key, param_name='api_key'):
        self.authenticator = ApiKeyAuthenticator(host=host,
                                                 api_key=api_key,
                                                 param_name=param_name)

    def set_token(self, host, token):
        self.authenticator = TokenAuthenticator(host=host, token=token)

    def set_jwt(self, host, header_name, jwt):
        self.authenticator = JwtAuthenticator(host, header_name, jwt)

    def set_nonce(self, host):
        self.authenticator = NonceAuthenticator(host)

    def request(self,
                method,
                url,
                params=None,
                data=None,
                headers=None,
                files=None,
                proxies=None):
        """Requests based implementation.

        :return: Requests response
        :rtype:  requests.Response
        """
        req = requests.Request(method=method,
                               url=url,
                               params=params,
                               data=data,
                               headers=headers,
                               files=files)
        self.apply_authentication(req)
        return self.session.send(self.session.prepare_request(req),
                                 verify=self.verify,
                                 proxies=proxies)

    def ws_connect(self, url, params=None):
        """Websocket-client based implementation.

        :return: WebSocket connection
        :rtype:  websocket.WebSocket
        """
        # Build a prototype request and apply authentication to it
        proto_req = requests.Request('GET', url, params=params)
        self.apply_authentication(proto_req)
        # Prepare the request, so params will be put on the url,
        # and authenticators can manipulate headers
        preped_req = proto_req.prepare()
        # Pull the Authorization header, if needed
        header = [
            "%s: %s" % (k, v) for (k, v) in list(preped_req.headers.items())
            if k == 'Authorization'
        ]
        # Pull the URL, which includes query params
        url = preped_req.url
        return websocket.create_connection(url, header=header)

    def apply_authentication(self, req):
        if self.authenticator and self.authenticator.matches(req.url):
            self.authenticator.apply(req)



#!/usr/bin/env python

#
# Copyright (c) 2013, Digium, Inc.
#
"""Main entry point for codegen command line app.
"""

import sys

from optparse import OptionParser

USAGE = "usage: %prog [options] template-dir output-dir"


def main(argv=None):
    """Main method, as invoked by setuptools launcher script.

    :param argv: Command line argument list.
    """
    if argv is None:
        argv = sys.argv

    parser = OptionParser(usage=USAGE)
    parser.add_option("-v",
                      "--verbose",
                      action="store_true",
                      dest="verbose",
                      default=False,
                      help="Verbose output")

    (options, args) = parser.parse_args(argv)

    if len(args) < 3:
        parser.error("Missing arguments")
    elif len(args) > 3:
        parser.error("Too many arguments")

    template_dir = args[1]
    output_dir = args[2]


# And sometimes you just want to run the script...
if __name__ == "__main__":
    sys.exit(main() or 0)



"""General utility functions
"""
import socket
import logging
import petname
import os
import requests
import socket
import xml.etree.ElementTree as ElementTree

from functools import wraps

logger = logging.getLogger(__name__)
logging.getLogger(__name__).setLevel(
    os.environ.get('TAPISPY_LOG_LEVEL', logging.WARNING))

__all__ = [
    'AttrDict', 'json_response', 'with_refresh', 'clients_url',
    'random_client_name'
]


class AttrDict(dict):
    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value


def json_response(f):
    @wraps(f)
    def _f(*args, **kwargs):
        resp = f(*args, **kwargs)
        resp.raise_for_status()
        return resp.json()

    return _f


def with_refresh(client, f, *args, **kwargs):
    """Call function ``f`` and refresh token if needed."""
    logger.debug('with_refresh()...')
    try:
        return f(*args, **kwargs)
    except requests.exceptions.HTTPError as exc:
        logger.debug('HTTPError: {0}'.format(exc))
        try:
            # Old versions of APIM return errors in XML:
            code = ElementTree.fromstring(exc.response.text)[0].text
            # raise SystemError(code)
        except Exception:
            # Any error here means the response was not XML.
            try:
                # Try to see if it's a json response,
                exc_json = exc.response.json()
                # if so, check if it is an expired token error (new versions of APIM return JSON errors):
                if "Invalid Credentials" in exc_json.get("fault").get(
                        "message"):
                    logger.info('client.refresh() and retry...')
                    #client.token.refresh()
                    client.refresh()
                    return f(*args, **kwargs)
                #  otherwise, return the JSON
                return exc.response
            except Exception:
                # Re-raise it, as it's not an expired token
                raise exc
        # only catch 'token expired' exception
        # other codes may mean a different error
        if code not in ["900903", "900904"]:
            logger.error('Unexpected error code: {0}'.format(code))
            raise

        if client.can_refresh:
            logger.info('Fallback: client.token.refresh() and retry...')
            client.refresh()

        return f(*args, **kwargs)


def clients_url(tenant_url):
    """Returns the clients API endpoint for a given tenant
    """
    return '{0}/clients/v2'.format(tenant_url)


def random_client_name(words=3, letters=6, hostname=False):
    """Generate a pseudorandom but human-readable Oauth client name
    """
    client_name = petname.generate(words=words, letters=letters)
    if hostname:
        client_name = socket.gethostname().lower() + '-' + client_name
    return client_name



from __future__ import absolute_import
import dateutil.parser
import json
import logging
import os
import requests
import sys
import time
import xml.etree.ElementTree as ElementTree

from future import standard_library
standard_library.install_aliases()  # noqa
from urllib.parse import urlsplit  # noqa
from urllib.request import getproxies  # noqa
from urllib.error import HTTPError  # noqa

from agavepy import settings

from agavepy.constants import (CACHES_DOT_DIR, AGPY_FILENAME, CACHE_FILENAME,
                               SESSIONS_FILENAME, TOKEN_SCOPE, TOKEN_TTL,
                               ENV_BASE_URL, ENV_TOKEN, ENV_REFRESH_TOKEN,
                               ENV_USERNAME, ENV_PASSWORD, ENV_API_KEY,
                               ENV_API_SECRET, ENV_TENANT_ID,
                               DEFAULT_TENANT_API_SERVER)

from agavepy.aloe import (LAST_PRE_ALOE_VERSION, EXCEPTION_MODELS)
from agavepy.configgen import (ConfigGen, load_resource)
from agavepy.errors import (AgaveError, AgaveException, __handle_tapis_error,
                            _handle_tapis_error)
from agavepy.interactive import ClientCommands, DeprecatedCommands, TokenCommands
from agavepy.processor import (AgaveProcessor, SwaggerClient, SwaggerClient,
                               SynchronousHttpClient, Operation, Resource)
from agavepy.tenants import id_by_api_server
from agavepy.token import Token
from agavepy.util import AttrDict, json_response, with_refresh

HERE = os.path.dirname(os.path.abspath(__file__))

logger = logging.getLogger(__name__)
logging.getLogger(__name__).setLevel(
    os.environ.get('TAPISPY_LOG_LEVEL', logging.WARNING))

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

__all__ = ['Agave']


class Agave(ClientCommands, TokenCommands, DeprecatedCommands):

    can_refresh = True

    PARAMS = [
        # param name, mandatory?, attr_name, default
        ("username", False, "username", None),
        ("password", False, "password", None),
        ("token_username", False, "token_username", None),
        ("jwt", False, "jwt", None),
        ("jwt_header_name", False, "header_name", None),
        ("api_server", False, "api_server", DEFAULT_TENANT_API_SERVER),
        ("tenant_id", False, "tenant_id", None),
        ("expires_in", False, "expires_in", None),
        ("expires_at", False, "expires_at", None),
        ("created_at", False, "created_at", None),
        ("client_name", False, "client_name", None),
        ("api_key", False, "api_key", None),
        ("api_secret", False, "api_secret", None),
        ("token", False, "_token", None),
        ("refresh_token", False, "_refresh_token", None),
        ("use_nonce", False, "use_nonce", False),
        ("resources", False, "resources", None),
        ("verify", False, "verify", True),
        ("token_callback", False, "token_callback", None),
        ("proxies", False, "proxies", getproxies()),
    ]

    def __init__(self, **kwargs):

        self.show_curl = settings.SHOW_CURL

        for param, mandatory, attr, default in self.PARAMS:
            try:
                value = (kwargs[param] if mandatory else kwargs.get(
                    param, default))
            except KeyError:
                raise AgaveError('parameter "{}" is mandatory'.format(param))
            setattr(self, attr, value)

        # Mark the client as unrefreshable if api server and token (but
        # not refresh_token) are provided, as it is considered a 'bare' client
        if self.api_key is None and self.api_secret is None:
            if self.api_server is not None and self._token is not None and self._refresh_token is None:
                setattr(self, 'can_refresh', False)

        if getattr(self, 'resources', None) is None:
            self.resources = load_resource(self.api_server)
        self.resource_exceptions = json.load(
            open(os.path.join(HERE, 'resource_exceptions.json'), 'r'))
        self.host = urlsplit(self.api_server).netloc
        if self.token_callback and not hasattr(self.token_callback,
                                               "__call__"):
            raise AgaveError('token_callback must be callable.')

        # If we are passed a JWT directly, we can bypass OAuth-related tasks
        if self.jwt:
            if not self.header_name:
                raise AgaveError(
                    'The jwt header name is required to use the jwt authenticator.'
                )

        self.token = None
        if (self.api_key is not None and self.api_secret is not None
                and self.jwt is None):
            self.set_client(self.api_key, self.api_secret)
        self.clients_resource = None
        self.all = None
        self.refresh_aris()

        if not hasattr(self, "apps"):
            raise AgaveError(
                "Required parameters for client instantiation missing.")

        logger.debug('Agave: {0}'.format(self))

    def to_dict(self):
        """Return a dictionary representing this client."""
        d = {}
        if hasattr(self, "token") and hasattr(self.token, "token_info"):
            d = {
                "token": self.token.token_info.get("access_token"),
                "refresh_token": self.token.token_info.get("refresh_token"),
                "expires_in": self.token.token_info.get("expires_in"),
                "expires_at": self.token.token_info.get("expires_at"),
                "created_at": self.token.token_info.get("created_at"),
            }
        d.update({
            attr: getattr(self, attr)
            for _, _, attr, _ in self.PARAMS if attr not in [
                "resources",
                "_token",
                "_refresh_token",
                "header_name",
                "jwt",
                "password",
                "token_callback",
            ]
        })
        # If we are writing to the .agave/current file, then
        # modify a few key fields accordingly
        if Agave.agpy_path() == Agave.tapis_current_path():
            d["tenantid"] = d.pop("tenant_id", "")
            d["apisecret"] = d.pop("api_secret", "")
            d["apikey"] = d.pop("api_key", "")
            d["baseurl"] = d.pop("api_server", "")
            d["access_token"] = d.pop("token", "")
        return d

    @classmethod
    def tapis_cache_path(self):
        """Return the path to the credentials cache directory
        """
        agcachedir = os.environ.get(
            "TAPIS_CACHE_DIR",
            os.environ.get("AGAVE_CACHE_DIR",
                           os.path.expanduser(
                               "~/{0}/".format(CACHES_DOT_DIR))))
        if os.path.isdir(agcachedir):
            return agcachedir
        elif not os.path.exists(agcachedir):
            os.makedirs(agcachedir)
            return agcachedir
        else:
            raise ValueError(
                "{0} does not appear to be a directory".format(agcachedir))

    @classmethod
    def tapis_current_path(self):
        """Return path to the current cached credential
        """
        agcachedir = self.tapis_cache_path()
        return os.path.join(agcachedir, CACHE_FILENAME)

    @classmethod
    def tapis_sessions_path(self):
        """Return path to the current sessions cache
        """
        agcachedir = self.tapis_cache_path()
        return os.path.join(agcachedir, SESSIONS_FILENAME)

    @classmethod
    def agpy_path(self):
        """Return path to an .agpy file
        """
        places = [
            self.tapis_current_path(),
            os.path.expanduser("~/{0}".format(AGPY_FILENAME)),
            "/etc/{0}".format(AGPY_FILENAME),
            "/root/{0}".format(AGPY_FILENAME),
            "/{0}".format(AGPY_FILENAME),
        ]
        for place in places:
            if os.path.exists(place):
                return place
        return places[0]

    @classmethod
    def _read_current(cls, agave_kwargs=False):
        with open(Agave.tapis_current_path()) as tcp:
            current = json.loads(tcp.read())
            if agave_kwargs:
                current['tenant_id'] = current.pop('tenantid', '')
                current['api_secret'] = current.pop('apisecret', '')
                current['api_key'] = current.pop('apikey', '')
                current['api_server'] = current.pop('baseurl', '')
                current['token'] = current.pop('access_token', '')
        return current

    @classmethod
    def _read_clients(cls):
        """Read clients from the credential cache file."""
        logger.debug('Agave._read_clients()...')
        with open(Agave.agpy_path()) as agpy:
            clients = json.loads(agpy.read())
        # If we are writing to the .agave/current file, then
        # translate a few key fields accordingly
        if Agave.agpy_path() == Agave.tapis_current_path():
            logger.debug('Using "current" cache format')
            # First, make sure we have a list; The BASH CLI uses a single
            # JSON object but .agpy can store multiple clients in a list
            if not isinstance(clients, list):
                clients = [clients]
            for client in clients:
                # convert CLI keys to agavepy keys:
                for k, v in list(client.items()):
                    if k == "tenantid":
                        client["tenant_id"] = v
                    elif k == "apisecret":
                        client["api_secret"] = v
                    elif k == "apikey":
                        client["api_key"] = v
                    elif k == "baseurl":
                        client["api_server"] = v
                    elif k == "access_token":
                        client["token"] = v
                # add missing attrs:
                if "verify" not in client:
                    # default to verifying SSL:
                    client["verify"] = True
                logger.debug('Found client {0}'.format(client))

        logger.debug('Total clients found: {0}'.format(len(clients)))
        return clients

    @classmethod
    def _read_sessions(cls):
        """Read sessions from the credentials cache."""
        logger.debug('Agave._read_sessions()...')
        try:
            with open(Agave.tapis_sessions_path()) as agsess:
                sessions = json.loads(agsess.read())
        except Exception:
            sessions = {'sessions': {}}
        return sessions

    @classmethod
    def _restore_direct(cls, api_server=None, token=None):
        """Initialize a non-refreshable client from server and token
        """
        if api_server is None:
            api_server = os.environ.get(ENV_BASE_URL, None)
        if token is None:
            token = os.environ.get(ENV_TOKEN, None)
        if api_server is not None and token is not None:
            ag = Agave(api_server=api_server, token=token)
            ag.can_refresh = False
            return ag
        else:
            raise AgaveError(
                'Either api_server or token cannot be resolved from parameters or environment'
            )

    @classmethod
    def _restore_cached(cls, **kwargs):
        """Restore a cached client from a specific attr."""

        try:
            clients = Agave._read_clients()
            # Deal with empty clients file
            if len(clients) == 0:
                logger.error('No clients found in cache.')
                raise AgaveError('No clients found in cache.')
            # If no attribute was passed, restore the first client in the list:
            if len(list(kwargs.items())) == 0:
                return Agave(**clients[0])
            else:
                for k, v in list(kwargs.items()):
                    for client in clients:
                        if client.get(k) == v:
                            return Agave(**client)
            raise AgaveError('Unable to resolve client by keyword')
        except FileNotFoundError:
            raise AgaveError('No cached credentials found')
        except Exception:
            raise

    @classmethod
    def _restore_env(cls):
        # Three types of client can be initialized from env variables
        #
        # 1. refresh is at least able to refresh a token but may also manage clients
        # and issue a new token pair if basic credentials are also provided
        # 2. basic client is only able to interact with clients and issue an initial token
        # 3. bare client is unable to refresh token or manage clients

        clients = {
            'refresh': [(ENV_BASE_URL, 'api_server', True),
                        (ENV_API_KEY, 'api_key', True),
                        (ENV_API_SECRET, 'api_secret', True),
                        (ENV_TOKEN, 'token', True),
                        (ENV_REFRESH_TOKEN, 'refresh_token', True),
                        (ENV_USERNAME, 'username', False),
                        (ENV_PASSWORD, 'password', False)],
            'basic': [(ENV_BASE_URL, 'api_server', True),
                      (ENV_USERNAME, 'username', True),
                      (ENV_PASSWORD, 'password', True),
                      (ENV_API_SECRET, 'api_secret', False),
                      (ENV_API_KEY, 'api_key', False)],
            'bare': [(ENV_BASE_URL, 'api_server', True),
                     (ENV_TOKEN, 'token', True)],
        }

        for kind, envkeys in clients.items():
            try:
                client = {}
                for env, key, req in envkeys:
                    val = os.environ.get(env, None)
                    if val is None and req is True:
                        raise KeyError('{0} must be present')
                    else:
                        client[key] = val
                logger.debug('Client-Type: {}'.format(kind))
                logger.debug('Client: {0}'.format(client))
                return Agave(**client)
            except Exception:
                pass

        logger.warning('restore_env has not returned yet')
        raise AgaveError(
            'One or more missing keys prevent loading a client from the environment'
        )

    @classmethod
    def restore(cls,
                api_key=None,
                client_name=None,
                tenant_id=None,
                token=None,
                api_server=None,
                cache_client=True):
        """Public API to load an Agave client from an external source
        """

        # Attempt bare client first
        try:
            logger.debug('Restore from api_server & token')
            return Agave._restore_direct(api_server=api_server, token=token)
        except AgaveError:
            pass

        # Attempt to load from cache file
        # (works with current, .agpy, and config.json formats)
        try:
            logger.debug('Restore from cached...')
            ag_from_cache = None

            # Resolve from cache by API key
            if api_key:
                logger.debug('By key {0}'.format(api_key))
                ag_from_cache = Agave._restore_cached(api_key=api_key)
            # Resolve from cache by client name
            elif client_name:
                logger.debug('By name {0}'.format(client_name))
                ag_from_cache = Agave._restore_cached(client_name=client_name)
            # Resolve from cache by tenant
            elif tenant_id:
                logger.debug('By tenant {0}'.format(tenant_id))
                ag_from_cache = Agave._restore_cached(tenant_id=tenant_id)
            else:
                # Load up default client
                logger.debug('By default client...')
                ag_from_cache = Agave._restore_cached()

            # Return client, after writing to session file
            if ag_from_cache is not None:
                if cache_client:
                    ag_from_cache._write_client()
                return ag_from_cache

        except FileNotFoundError:
            raise
        except AgaveError:
            pass

        # Attempt to load from environment
        try:
            logger.debug('Restore from env')
            ag = Agave._restore_env()
            if cache_client:
                try:
                    ag._write_client()
                except Exception:
                    raise
            return ag
        except AgaveError:
            pass

        raise AgaveError(
            'Unable to restore a client from parameters, cache, or environment'
        )

    def _write_client(self, permissive=True):
        """Update the credential and sessions cache with the latest values
        """
        # If we are working with an 'current' file, then
        # use the older, less descriptive JSON format
        logger.debug('Agave._write_client()...')
        logger.debug('Target: {0}'.format(Agave.tapis_current_path()))

        if self.can_refresh is False:
            if permissive is False:
                raise NotImplementedError(
                    'Cannot cache credentials if client is not configured for token refresh'
                )
            else:
                return False

        if Agave.agpy_path() == Agave.tapis_current_path():
            logger.debug('writing "current" cache file...')
            new_data = self.to_dict()

            try:
                with open(Agave.agpy_path(), "r") as agpy:
                    old_data = json.loads(agpy.read())
            except Exception:
                old_data = new_data

            # Interesting - we don't seem to carry new_data's time and expiry
            # when writing the cache file. Could that be causing issues w/
            # consumer applications that rely on the cache ttl to implement
            # refresh logic?
            old_data["access_token"] = new_data["access_token"]
            old_data["refresh_token"] = new_data["refresh_token"]

            # Some Tapis client managers drop tenant_id - this fixes that issue
            if getattr(self, 'tenant_id', None) is None:
                self.tenant_id = id_by_api_server(self.api_server)
                new_data['tenantid'] = self.tenant_id

            with open(Agave.agpy_path(), "w") as agpy:
                agpy.write(json.dumps(new_data))
            logger.debug('"current" cache file written')

            try:
                # Write to sessions sidecar
                logger.debug('writing "sessions" cache...')
                sessions = Agave._read_sessions()
                sessions_data = sessions["sessions"]

                tenant_id = new_data["tenantid"]
                username = new_data["username"]

                current_name = new_data.get("client_name", None)
                if current_name is None:
                    current_name = new_data.get("apikey", None)

                if tenant_id not in sessions_data:
                    sessions_data[tenant_id] = {}

                if username not in sessions_data[tenant_id]:
                    sessions_data[tenant_id][username] = {}
                # if current_name not in sessions_data[tenant_id][username]
                # Should probably do an update or merge here
                sessions_data[tenant_id][username][current_name] = new_data
                sessions["sessions"] = sessions_data
                with open(Agave.tapis_sessions_path(), "w") as agsess:
                    agsess.write(json.dumps(sessions))
                logger.debug('"sessions" cache written')

            except Exception as exc:
                # Failure to write sessions file should never cause a failure
                logger.error('failed writing sessions file: {0}'.format(exc))

        else:
            clients = Agave._read_clients()
            new_clients = []
            for client in clients:
                # Update client with the latest representation
                if (client.get("api_key") == self.api_key
                        or client.get("client_name") == self.client_name):
                    new_clients.append(self.to_dict())
                else:
                    # otherwise, write what is already there.
                    new_clients.append(client)
            with open(Agave.agpy_path(), "w") as agpy:
                agpy.write(json.dumps(new_clients))

    def refresh_aris(self):
        logger.debug('Agave.refresh_aris()...')
        self.clients_ari()
        # the resources are defined with a different authenticator in case
        # a jwt is passed in, hence we need to refresh using a different method
        if self.jwt:
            self.jwt_ari()
        elif self.use_nonce:
            self.nonce_ari()
        else:
            self.full_ari()

    def clients_ari(self):
        # If there is enough information to establish HTTP basic auth,
        # then create a 'clients' resource object
        self.clients_resource = self.resource("basic_auth", "host", "username",
                                              "password")

    def full_ari(self):
        # If there is a token, then create a resource object with a
        # bearer token
        self.all = self.resource("token", "host", "_token")

    def nonce_ari(self):
        # if use_nonce is passed in, we use the NonceAuthenticator which does
        # not require any credentials the time of instantiation.
        self.all = self.resource("nonce", "host")

    def jwt_ari(self):
        # If a jwt is passed in, create a resource object with
        # the header_name and jwt token:
        self.all = self.resource("jwt", "host", "header_name", "jwt")

    def resource(self, auth_type, *args):
        logger.debug('Agave.resource({0})...'.format(auth_type))
        args_values = [getattr(self, arg) for arg in args]
        if all(args_values):
            http_client = SynchronousHttpClient(verify=self.verify)
            auth = getattr(http_client, "set_{}".format(auth_type))
            # auth method will be one of: set_basic_auth, set_token,
            # or set_jwt depending on params passed to the constructor.
            auth(*args_values)
            return SwaggerClient(self.resources,
                                 http_client=http_client,
                                 extra_processors=[AgaveProcessor()],
                                 show_curl=self.show_curl)

    def set_client(self, key, secret):
        """
        :type key: str
        :type secret: str
        :rtype: None
        """
        logger.debug('Agave.set_client({0}, {1})...'.format(key, '******'))

        self.api_key = key
        self.api_secret = secret
        self.token = Token(
            self.username,
            self.password,
            self.api_server,
            self.api_key,
            self.api_secret,
            self.verify,
            self,
            self._token,
            self._refresh_token,
            self.token_username,
            self.expires_at,
            self.expires_in,
            self.created_at,
        )
        if self._token:
            logger.debug('self._token exists')
            pass
        else:
            logger.debug('requesting token creation...')
            self.token.create()
        self.refresh_aris()
        logger.debug('set_client() complete')

    def geturl(self, url):
        """Make get request to url using the client access token and retry if token has expired.
        :param url: str
        :return:
        """
        f = requests.get
        return with_refresh(
            self.client,
            f,
            url,
            headers={"Authorization": "Bearer " + self._token},
            verify=self.verify,
            proxies=self.proxies,
        )

    def download_uri(self, uri, local_path):
        """Convenience method to download an agave URL or jobs output URL to an
        absolute `path` on the local file system."""
        if uri.startswith("http") and "jobs" in uri:
            # assume job output uri:
            if "/outputs/listings/" in uri:
                download_url = uri.replace("listings", "media")
            elif "/outputs/media/" in uri:
                download_url = uri
            else:
                raise AgaveError("Unsupported jobs URI.")
        elif "agave://" in uri:
            # assume it is an agave uri
            system_id, path = uri.split("agave://")[1].split("/", 1)
            download_url = "{}/files/v2/media/system/{}/{}".format(
                self.api_server, system_id, path)
        else:
            raise AgaveError("Unsupported URI.")
        f = requests.get
        with open(local_path, "wb") as loc:
            rsp = with_refresh(
                self.client,
                f,
                download_url,
                headers={"Authorization": "Bearer " + self._token},
                verify=self.verify,
                proxies=self.proxies,
            )
            rsp.raise_for_status()
            if type(rsp) == dict:
                raise AgaveError(
                    "Error downloading file at URI: {}, Response: {}".format(
                        uri, rsp))
            for block in rsp.iter_content(1024):
                if not block:
                    break
                loc.write(block)

    def __getattr__(self, key):
        return Resource(key, client=self)

    def __dir__(self):
        base = []
        if hasattr(self, "clients_resource") and hasattr(
                self.clients_resource, "resources"):
            base.extend(list(self.clients_resource.resources.keys()))
        if self.all is not None:
            base.extend(list(self.all.resources.keys()))
        return list(set(base))



"""Token management
"""

import logging
import os
import requests
import time

from future import standard_library
standard_library.install_aliases()  # noqa
from urllib.parse import urlparse, urlencode, urljoin, urlsplit  # noqa

from agavepy.constants import (CACHES_DOT_DIR, AGPY_FILENAME, CACHE_FILENAME,
                               SESSIONS_FILENAME, TOKEN_SCOPE, TOKEN_TTL,
                               ENV_BASE_URL, ENV_TOKEN, ENV_REFRESH_TOKEN,
                               ENV_USERNAME, ENV_PASSWORD, ENV_API_KEY,
                               ENV_API_SECRET, ENV_TENANT_ID)

logger = logging.getLogger(__name__)
logging.getLogger(__name__).setLevel(
    os.environ.get('TAPISPY_LOG_LEVEL', logging.WARNING))

__all__ = ['Token']


class Token(object):
    def __init__(
            self,
            username,
            password,
            api_server,
            api_key,
            api_secret,
            verify,
            parent,
            _token=None,
            _refresh_token=None,
            token_username=None,
            expires_at=None,
            expires_in=None,
            created_at=None,
    ):
        self.username = username
        self.password = password
        self.api_server = api_server
        self.api_key = api_key
        self.api_secret = api_secret
        self.token_username = token_username
        # Agave object that created this token
        self.parent = parent
        self.verify = verify
        if _token and _refresh_token:
            self.token_info = {
                "access_token": _token,
                "refresh_token": _refresh_token
            }
            self.token_info["expires_at"] = expires_at
            self.token_info["expires_in"] = expires_in
            self.token_info["created_at"] = created_at
            self.parent._token = _token

        self.token_url = urljoin(str(self.api_server), "token")
        logger.debug('Token URL: {0}'.format(self.token_url))

    def _token(self, data):
        logger.debug('Token._token({0})...'.format(data))
        auth = requests.auth.HTTPBasicAuth(self.api_key, self.api_secret)
        resp = requests.post(
            self.token_url,
            data=data,
            auth=auth,
            verify=self.verify,
            proxies=self.parent.proxies,
        )
        resp.raise_for_status()
        self.token_info = resp.json()
        logger.debug('self.token_info: {0}'.format(self.token_info))

        # The following enforces a maximum *reported* token TTL
        #
        # For clients that consult the expiration data in the auth
        # cache to determine whether to refresh, this will solve most of
        # their credential expiry issues
        expires_in = int(self.token_info.get("expires_in", TOKEN_TTL))
        if expires_in > TOKEN_TTL:
            expires_in = TOKEN_TTL
        logger.debug('expires_in: {0} seconds'.format(expires_in))

        created_at = int(time.time())
        self.token_info["expires_in"] = expires_in
        self.token_info["created_at"] = created_at
        self.token_info["expiration"] = created_at + expires_in
        self.token_info["expires_at"] = time.ctime(created_at + expires_in)
        token = self.token_info["access_token"]
        logger.debug('token: {0}'.format(token))

        # Update parent with new token data
        self.parent._token = token
        self.parent.refresh_token = self.token_info["refresh_token"]
        self.parent.created_at = self.token_info["created_at"]
        self.parent.expiration = self.token_info["expiration"]
        self.parent.expires_at = self.token_info["expires_at"]
        self.parent.expires_in = self.token_info["expires_in"]

        # try to persist the token data
        try:
            self.parent._write_client()
        except Exception as exc:
            # failing to writing the cache file cannot block use.
            logger.warning('Failed to write client: {0}'.format(exc))
            pass

        if self.parent.token_callback:
            self.parent.token_callback(**self.token_info)
        self.parent.refresh_aris()
        return token

    def create(self):
        logger.debug('Token.create()...')
        data = {
            "grant_type": "password",
            "username": self.username,
            "password": self.password,
            "scope": TOKEN_SCOPE,
        }
        if self.token_username:
            data["grant_type"] = "admin_password"
            data["token_username"] = self.token_username

        return self._token(data)

    def refresh(self):
        logger.debug('Token.refresh()...')
        data = {
            "grant_type": "refresh_token",
            "scope": TOKEN_SCOPE,
            "refresh_token": self.token_info["refresh_token"],
        }

        logger.debug('Token.refresh() finished')
        return self._token(data)



import jinja2
import json
import logging
import os

from future import standard_library
standard_library.install_aliases()  # noqa
from urllib.parse import urlparse, urlencode, urljoin, urlsplit  # noqa

HERE = os.path.dirname(os.path.abspath(__file__))

logger = logging.getLogger(__name__)
logging.getLogger(__name__).setLevel(
    os.environ.get('TAPISPY_LOG_LEVEL', logging.WARNING))

__all__ = ['ConfigGen', 'load_resource', 'updateDict']


class ConfigGen(object):
    def __init__(self, template_str):
        self.template_str = template_str

    def compile(self, configs, env):
        template = env.get_template(self.template_str)
        return template.render(configs)


def updateDict(base_dict, new_dict):
    for key, val in new_dict.items():
        if isinstance(val, dict):
            base_dict[key] = updateDict(base_dict.get(key, {}), val)
        elif isinstance(val, list):
            if key not in base_dict:
                base_dict[key] = []
            for list_dict in val:
                base_dict[key].append(list_dict)
        else:
            base_dict[key] = val
    return base_dict


def load_resource(api_server):
    """Load a default resource file.

    :type api_server: str
    :rtype: dict
    """
    logger.debug('load_resource({0})...'.format(api_server))
    rsrcs = {}
    rsrc_files = [
        "resources/misc.json.j2",
        "resources/api_clients.json.j2",
        "resources/api_apps.json.j2",
        "resources/api_files.json.j2",
        "resources/api_jobs.json.j2",
        "resources/api_meta.json.j2",
        "resources/api_monitors.json.j2",
        "resources/api_notifications.json.j2",
        "resources/api_postits.json.j2",
        "resources/api_profiles.json.j2",
        "resources/api_systems.json.j2",
        "resources/api_transforms.json.j2",
        "resources/api_actors.json.j2",
        "resources/api_admin.json.j2",
    ]
    for rsrc_file in rsrc_files:
        conf = ConfigGen(rsrc_file)
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(HERE),
                                 trim_blocks=True,
                                 lstrip_blocks=True)

        new_rsrcs = json.loads(
            conf.compile({"api_server_base": urlparse(api_server).netloc},
                         env))
        updateDict(rsrcs, new_rsrcs)

    logger.debug('load_resource finished')
    return rsrcs



import json
from .agave import Agave


def switch(client_name_or_key, username=None, tenant_id=None):
    # Read in current config
    try:
        current = Agave._read_current()
        # Try to inherit username and tenant from 'current' if not provided
        # This will let single-tenant switchers worry only about client_name
        if username is None:
            username = current.get('username')
        if tenant_id is None:
            tenant_id = current.get('tenantid')
    except IndexError:
        current = None

    # Read in sessions, then see if we can match to passed params
    try:
        complete_sessions = Agave._read_sessions()
        sessions = complete_sessions.get('sessions')
        if sessions is None:
            raise ValueError()
    except Exception:
        sessions = dict()
        sessions[tenant_id] = {}
        sessions[tenant_id][username] = {}
        complete_sessions = {'current': None, 'sessions': sessions}

    def resolve_client(sess, tenant_id, username, client_name_or_key):
        for k, v in sess.get(tenant_id, {}).get(username, {}).items():
            if k == client_name_or_key:
                return v
            elif v.get('apikey', None) == client_name_or_key:
                # v['client_name'] = client_name_or_key
                return v
        raise KeyError('Unable to find a matching client')

    session = resolve_client(sessions, tenant_id, username, client_name_or_key)

    print(session)

    if session.get('client_name', None) == client_name_or_key:
        if session.get('username') == current.get('username', None):
            if session.get('tenantid') == current.get('tenantid'):
                print('Current and target sessions do not differ')
                return False

    # Replace current file
    with open(Agave.tapis_current_path(), 'w') as current_file:
        # If the client was created in a client name-aware manner,
        # ensure client_name stays associated with the 'current' file
        if 'client_name' not in session or session.get('client_name') == '':
            session['client_name'] = client_name_or_key
        current_file.write(json.dumps(session))

    # Write out sessions file
    if 'client_name' not in current or current.get('client_name') == '':
        current['client_name'] = client_name_or_key
    complete_sessions['current'] = session
    complete_sessions['sessions'][tenant_id][username][
        client_name_or_key] = current

    # Write updated sessions file
    with open(Agave.tapis_sessions_path(), 'w') as sessions_file:
        sessions_file.write(json.dumps(complete_sessions))

    # They are the same - no need to switch

    # # Fetch the client from sessions.json that matches the tenant and user,
    # # plus either the client_name or the current.apikey
    # try:
    #     try:
    #         target_client = clients.get('sessions', {}).get(tenant_id, {}).get(
    #             username, {}).get(client_name_or_key, None)
    #     except KeyError:
    #         target_client = resolve_client(clients.get(
    #             'sessions', {}), tenant_id, username, client_name_or_key)
    # except KeyError:
    #     # Client did not exist in sessions cache
    #     target_client = dict()

    # # Keys arent same
    # if client['apikey'] != target_client.get('apikey', None):
    #     pass



"""Provides environment variable-driven runtime configuration
"""
PKG_NAME = 'tapis_py'
import os
import warnings
from dotenv import load_dotenv, find_dotenv

with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    if not load_dotenv(find_dotenv()):
        if not load_dotenv(find_dotenv(usecwd=True)):
            load_dotenv(os.path.join(os.path.expanduser('~'), '.env'))

from attrdict import AttrDict  # noqa
from .helpers import ENV_PREFIX  # noqa
from .log import *  # noqa
from .tenancy import *  # noqa


def all_settings():
    """Returns name and value of all settings
    """
    from types import ModuleType

    settings = {}
    for name, item in globals().items():
        # Ignore callables and private properties
        if not callable(item) and not name.startswith("__") \
                and not isinstance(item, ModuleType):
            settings[name] = item
    return AttrDict(settings)



#
# Copyright (c) 2013, Digium, Inc.
#
"""Swagger processors enrich and validate the Swagger data model.

This can be to make templating easier, or ensure values required for a
particular use case (such as ensuring that description and summary fields
exist)
"""


class ParsingContext(object):
    """Context information for parsing.

    This object is immutable. To change contexts (like adding an item to the
    stack), use the next() and next_stack() functions to build a new one.
    """
    def __init__(self):
        self.type_stack = []
        self.id_stack = []
        self.args = {'context': self}

    def __repr__(self):
        zipped = list(zip(self.type_stack, self.id_stack))
        strs = ["%s=%s" % (t, i) for (t, i) in zipped]
        return "ParsingContext(stack=%r)" % strs

    def is_empty(self):
        """Tests whether context is empty.

        :return: True if empty, False otherwise.
        """
        return not self.type_stack and not self.id_stack

    def push(self, obj_type, json, id_field):
        """Pushes a new self-identifying object into the context.

        :type obj_type: str
        :param json: Specifies type of object json represents
        :type json: dict
        :param json: Current Jsonified object.
        :type id_field: str
        :param id_field: Field name in json that identifies it.
        """
        if id_field not in json:
            raise SwaggerError("Missing id_field: %s" % id_field, self)
        self.push_str(obj_type, json, str(json[id_field]))

    def push_str(self, obj_type, json, id_string):
        """Pushes a new object into the context.

        :type obj_type: str
        :param obj_type: Specifies type of object json represents
        :type json: dict
        :param json: Current Jsonified object.
        :type id_string: str
        :param id_string: Identifier of the given json.
        """
        self.type_stack.append(obj_type)
        self.id_stack.append(id_string)
        self.args[obj_type] = json

    def pop(self):
        """Pops the most recent object out of the context
        """
        del self.args[self.type_stack.pop()]
        self.id_stack.pop()


class SwaggerError(Exception):
    """Raised when an error is encountered mapping the JSON objects into the
    model.
    """
    def __init__(self, msg, context, cause=None):
        """Ctor.

        :param msg: String message for the error.
        :param context: ParsingContext object
        :param cause: Optional exception that caused this one.
        """
        super(Exception, self).__init__(msg, context, cause)


class SwaggerProcessor(object):
    """Post processing interface for Swagger API's.

    This processor can add fields to model objects for additional
    information to use in the templates.
    """
    def apply(self, resources):
        """Apply this processor to a loaded Swagger definition.

        :param resources: Top level Swagger definition.
        :type  resources: dict
        """
        context = ParsingContext()
        resources_url = resources.get('url') or 'json:resource_listing'
        context.push_str('resources', resources, resources_url)
        self.process_resource_listing(**context.args)
        for listing_api in resources['apis']:
            context.push('listing_api', listing_api, 'path')
            self.process_resource_listing_api(**context.args)
            context.pop()

            api_url = listing_api.get('url') or 'json:api_declaration'
            context.push_str('resource', listing_api['api_declaration'],
                             api_url)
            self.process_api_declaration(**context.args)
            for api in listing_api['api_declaration']['apis']:
                context.push('api', api, 'path')
                self.process_resource_api(**context.args)
                for operation in api['operations']:
                    context.push('operation', operation, 'nickname')
                    self.process_operation(**context.args)
                    for parameter in operation.get('parameters', []):
                        context.push('parameter', parameter, 'name')
                        self.process_parameter(**context.args)
                        context.pop()
                    for response in operation.get('errorResponses', []):
                        context.push('error_response', response, 'code')
                        self.process_error_response(**context.args)
                        context.pop()
                    context.pop()
                context.pop()
            models = listing_api['api_declaration'].get('models', {})
            for (name, model) in list(models.items()):
                context.push('model', model, 'id')
                self.process_model(**context.args)
                for (name, prop) in list(model['properties'].items()):
                    context.push('prop', prop, 'name')
                    self.process_property(**context.args)
                    context.pop()
                context.pop()
            context.pop()
        context.pop()
        assert context.is_empty(), "Expected %r to be empty" % context

    def process_resource_listing(self, resources, context):
        """Post process a resources.json object.

        :param resources: ResourceApi object.
        :type context: ParsingContext
        :param context: Current context in the API.
        """
        pass

    def process_resource_listing_api(self, resources, listing_api, context):
        """Post process entries in a resource.json's api array.

        :param resources: Resource listing object
        :param listing_api: ResourceApi object.
        :type context: ParsingContext
        :param context: Current context in the API.
        """
        pass

    def process_api_declaration(self, resources, resource, context):
        """Post process a resource object.

        This is parsed from a .json file reference by a resource listing's
        'api' array.

        :param resources: Resource listing object
        :param resource: resource object.
        :type context: ParsingContext
        :param context: Current context in the API.
        """
        pass

    def process_resource_api(self, resources, resource, api, context):
        """Post process entries in a resource's api array

        :param resources: Resource listing object
        :param resource: resource object.
        :param api: API object
        :type context: ParsingContext
        :param context: Current context in the API.
        """
        pass

    def process_operation(self, resources, resource, api, operation, context):
        """Post process an operation on an api.

        :param resources: Resource listing object
        :param resource: resource object.
        :param api: API object
        :param operation: Operation object.
        :type context: ParsingContext
        :param context: Current context in the API.
        """
        pass

    def process_parameter(self, resources, resource, api, operation, parameter,
                          context):
        """Post process a parameter on an operation.

        :param resources: Resource listing object
        :param resource: resource object.
        :param api: API object
        :param operation: Operation object.
        :param parameter: Parameter object.
        :type context: ParsingContext
        :param context: Current context in the API.
        """
        pass

    def process_error_response(self, resources, resource, api, operation,
                               error_response, context):
        """Post process an errorResponse on an operation.

        :param resources: Resource listing object
        :param resource: resource object.
        :param api: API object
        :param operation: Operation object.
        :param error_response: Response object.
        :type context: ParsingContext
        :param context: Current context in the API.
        """
        pass

    def process_model(self, resources, resource, model, context):
        """Post process a model from a resources model dictionary.

        :param resources: Resource listing object
        :param resource: resource object.
        :param model: Model object.
        :type context: ParsingContext
        :param context: Current context in the API.
        """
        pass

    def process_property(self, resources, resource, model, prop, context):
        """Post process a property from a model.

        :param resources: Resource listing object
        :param resource: resource object.
        :param model: Model object.
        :param prop: Property object.
        :type context: ParsingContext
        :param context: Current context in the API.
        """
        pass


# noinspection PyDocstring
class WebsocketProcessor(SwaggerProcessor):
    """Process the WebSocket extension for Swagger
    """
    def process_resource_api(self, resources, resource, api, context):
        api.setdefault('has_websocket', False)

    def process_operation(self, resources, resource, api, operation, context):
        operation['is_websocket'] = operation.get('upgrade') == 'websocket'

        if operation['is_websocket']:
            api['has_websocket'] = True
            if operation['httpMethod'] != 'GET':
                raise SwaggerError(
                    "upgrade: websocket is only valid on GET operations",
                    context)


# noinspection PyDocstring
class FlatenningProcessor(SwaggerProcessor):
    """Flattens model and property dictionaries into lists.

    Mustache requires a regular schema.
    """
    def process_api_declaration(self, resources, resource, context):
        resource.model_list = list(resource.models.values())

    def process_model(self, resources, resource, model, context):
        # Convert properties dict to list
        model.property_list = list(model.properties.values())



"""Public, low-dependency helper functions
"""
import arrow
import os
import pkg_resources
import six

from dateutil.parser import parse


def ts_to_isodate(date_string, include_time=False):
    """Convert a datetime string (UTC) into a date string in ISO format"""

    iso_date_str = date_string

    try:
        date = parse(date_string)
        if include_time:
            if six.PY3:
                iso_date_str = date.isoformat(timespec='seconds')
            else:
                iso_date_str = date.isoformat()
        else:
            iso_date_str = date.date().isoformat()
    except ValueError:
        pass

    return iso_date_str


def ts_to_date(date_string):
    """Convert a datetime string (UTC) into a pretty date string"""

    date_rep = date_string

    try:
        date = parse(date_string)
        date_rep = date.strftime('%b %d %H:%M:%S %Y')
    except ValueError:
        pass

    return date_rep


def datetime_to_isodate(date_obj):
    """Convert a Python datetime object to ISO-8601
    """
    return arrow.get(date_obj).format('YYYY-MM-DDTHH:mm:ss.SSSZZ')


def datetime_to_human(date_obj):
    """Convert a Python datetime object to a human-friendly string
    """
    return arrow.get(date_obj).humanize()


def humanize_bytes(bytesize, precision=2):
    """
    Humanize byte size figures
    """
    # Reference: https://gist.github.com/moird/3684595

    abbrevs = ((1 << 50, 'PB'), (1 << 40, 'TB'), (1 << 30, 'GB'),
               (1 << 20, 'MB'), (1 << 10, 'kB'), (1, 'bytes'))
    if bytesize == 1:
        return '1 byte'
    for factor, suffix in abbrevs:
        if bytesize >= factor:
            break
    if factor == 1:
        precision = 0
    return '%.*f %s' % (precision, bytesize / float(factor), suffix)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
from setuptools import setup
from setuptools.command.test import test as TestCommand

import os
import sys

HERE = os.path.dirname(__file__)


class PyTest(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['--ignore', 'build']
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

readme = open('README.rst').read()
requires = [pkg for pkg in open('requirements.txt').readlines()]

# append the python2 requirements
if sys.version_info[0] == 2:
    requires.extend([pkg for pkg in open('requirements-py2.txt').readlines()])

data_files = [('', ['requirements.txt', 'requirements-py2.txt'])]

setup(
    name='agavepy',
    version='1.0.0a4',
    description='SDK for TACC Tapis (formerly Agave)',
    long_description=readme,
    author='Texas Advanced Computing Center',
    author_email='jstubbs@tacc.utexas.edu, vaughn@tacc.utexas.edu',
    url='https://github.com/TACC/agavepy',
    packages=[
        'agavepy',
        'agavepy.interactive',
        'agavepy.swaggerpy',
        'agavepy.settings',
        'agavepy.tenants'
    ],
    package_dir={'agavepy': 'agavepy'},
    package_data={'agavepy': [
        'resources.json', 'resources.json.j2', 'resource_exceptions.json', 'resources/*.j2']},
    data_files=data_files,
    install_requires=requires,
    license="BSD",
    zip_safe=False,
    keywords='',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'

    ],
    cmdclass={'test': PyTest},
    tests_require=['pytest'],
    test_suite='tests',
)



"""Aloe compatibility
"""

__all__ = ['LAST_PRE_ALOE_VERSION', 'EXCEPTION_MODELS']

LAST_PRE_ALOE_VERSION = '2.2.22-r7deb380'

# response models which are dependent on the version of the API response -
EXCEPTION_MODELS = {
    "Job": {
        "version_cutoff": LAST_PRE_ALOE_VERSION,
        "model": "AloeJob"
    },
    "JobSummary": {
        "version_cutoff": LAST_PRE_ALOE_VERSION,
        "model": "AloeJobSummary"
    },
}



import os
from agavepy.constants import (ENV_TENANTS_URL, ENV_TENANT_ID,
                               DEFAULT_TENANTS_URL, DEFAULT_TENANT_ID,
                               ENV_DEFAULT_TENANT_ID)

__all__ = ['TAPIS_TENANTS_URL', 'TAPIS_DEFAULT_TENANT_ID']

TAPIS_TENANTS_URL = os.environ.get(ENV_TENANTS_URL, DEFAULT_TENANTS_URL)

TAPIS_DEFAULT_TENANT_ID = os.environ.get(ENV_DEFAULT_TENANT_ID,
                                         DEFAULT_TENANT_ID)



import requests
from agavepy.util import clients_url, random_client_name

__all__ = ['ClientCommands']


class ClientCommands(object):
    def clients_create(self,
                       client_name=None,
                       description=None,
                       tenant_url=None,
                       username=None,
                       password=None,
                       quiet=False):
        """ Create an Oauth client

        Make a request to the API to create an Oauth client. Returns the client
        API key and secret as a tuple.

        KEYWORD ARGUMENTS
        -----------------
        client_name: string
            Name for Oauth2 client.
        description: string
            Description of the Oauth2 client
        tenant_url: string
            URL of the API tenant to interact with
        username: string
            The user's username.
        password: string
            The user's password

        RETURNS
        -------
        api_key: string
        api_secret: string
        """

        # Set request endpoint.
        if tenant_url is None:
            tenant_url = getattr(self, 'api_server')
        endpoint = clients_url(tenant_url)

        # User credentials
        if username is None:
            username = getattr(self, 'username')
        if password is None:
            password = getattr(self, 'password')

        # Make sure client_name is not empty
        if client_name == '' or client_name is None:
            client_name = random_client_name(words=2, hostname=True)

        # Make request.
        try:
            data = {
                'clientName': client_name,
                'description': description,
                'tier': 'Unlimited',
                'callbackUrl': '',
            }
            response = requests.post(endpoint,
                                     data=data,
                                     auth=(username, password))
            del password
        except Exception:
            del password
            raise

        # Parse the request's response and return api key and secret.
        result = response.json().get('result', {})
        api_key = result.get('consumerKey')
        api_secret = result.get('consumerSecret')
        if api_key == '' or api_secret == '':
            raise requests.exceptions.HTTPError(
                'Failed to create client {0}'.format(client_name))

        return {
            'api_key': api_key,
            'api_secret': api_secret,
            'client_name': client_name
        }



"""Stub support for 0.9.x experimental interactive commands
"""
import requests
from agavepy.util import clients_url, random_client_name

DISCONTINUED_MSG = 'This function is discontinued. Consult the Tapis CLI Python API for an analogous capability.'

__all__ = ['DiscontinuedError', 'DeprecatedCommands']


class DiscontinuedError(NotImplementedError):
    def __init__(self):
        NotImplementedError.__init__(self, DISCONTINUED_MSG)


class DeprecatedCommands(object):
    def init(self, tenantsurl=None):
        raise DiscontinuedError

    def save_configs(self, cache_dir=None):
        raise DiscontinuedError

    def load_configs(self,
                     cache_dir=None,
                     tenant_id=None,
                     username=None,
                     client_name=None):
        raise DiscontinuedError

    def load_client(self,
                    cache_dir=None,
                    tenant_id=None,
                    username=None,
                    client_name=None):
        raise DiscontinuedError

    def list_tenants(self, tenantsurl=None):
        raise DiscontinuedError

    def clients_delete(self,
                       client_name=None,
                       username=None,
                       password=None,
                       quiet=False):
        raise DiscontinuedError

    def clients_subscribe(self,
                          api_name,
                          api_version,
                          api_provider,
                          client_name=None,
                          username=None,
                          password=None,
                          quiet=False):
        raise DiscontinuedError

    def clients_subscriptions(self,
                              client_name=None,
                              username=None,
                              password=None,
                              quiet=False):
        raise DiscontinuedError

    def clients_list(self, username=None, password=None, quiet=False):
        raise DiscontinuedError

    def get_access_token(self, username=None, password=None, quiet=False):
        raise DiscontinuedError

    def refresh_tokens(self, force=False):
        raise DiscontinuedError

    def files_copy(self, source, destination):
        raise DiscontinuedError

    def files_delete(self, file_path):
        raise DiscontinuedError

    def files_download(self, source, destination):
        raise DiscontinuedError

    def files_history(self, path):
        raise DiscontinuedError

    def files_import(self, source, destination):
        raise DiscontinuedError

    def files_list(self, system_path, long_format=False):
        raise DiscontinuedError

    def files_mkdir(self, location):
        raise DiscontinuedError

    def files_move(self, source, destination):
        raise DiscontinuedError

    def files_pems_delete(self, path):
        raise DiscontinuedError

    def files_pems_list(self, path):
        raise DiscontinuedError

    def files_pems_update(self, path, username, perms, recursive=False):
        raise DiscontinuedError

    def files_upload(self, source, destination):
        raise DiscontinuedError



"""Global constants for the AgavePy package
"""
__all__ = [
    'PLATFORM', 'BASE_URL', 'SESSION_CACHE_DIRS', 'CACHE_FILENAME',
    'SESSIONS_FILENAME', 'AGPY_FILENAME', 'ENV_USERNAME', 'ENV_PASSWORD',
    'ENV_TOKEN', 'ENV_REFRESH_TOKEN', 'ENV_BASE_URL', 'ENV_API_KEY',
    'ENV_API_SECRET', 'ENV_TENANT_ID', 'TOKEN_SCOPE', 'TOKEN_TTL',
    'ENV_TENANTS_URL', 'DEFAULT_TENANTS_URL', 'DEFAULT_TENANT_ID',
    'ENV_DEFAULT_TENANT_ID', 'DEFAULT_TENANT_API_SERVER'
]

# General
PLATFORM = 'Tapis'
BASE_URL = 'https://api.tacc.utexas.edu'
DEFAULT_TENANTS_URL = BASE_URL + '/tenants'
DEFAULT_TENANT_ID = 'tacc.prod'
DEFAULT_TENANT_API_SERVER = BASE_URL

# Variable names usable in lieu of passed values to allow environment-driven
# configuration of the library
ENV_USERNAME = 'TAPIS_USERNAME'
ENV_PASSWORD = 'TAPIS_PASSWORD'
ENV_TOKEN = 'TAPIS_TOKEN'
ENV_REFRESH_TOKEN = 'TAPIS_REFRESH_TOKEN'
ENV_BASE_URL = 'TAPIS_BASE_URL'
ENV_TENANT_ID = 'TAPIS_TENANT_ID'
ENV_API_KEY = 'TAPIS_API_KEY'
ENV_API_SECRET = 'TAPIS_API_SECRET'
ENV_TENANTS_URL = 'TAPIS_TENANTS_URL'
ENV_DEFAULT_TENANT_ID = 'TAPIS_DEFAULT_TENANT_ID'

# Cache and session maangement
SESSION_CACHE_DIRS = ('TAPIS_CACHE_DIR', 'TAPIS_CACHE', 'AGAVE_CACHE_DIR')
CACHES_DOT_DIR = '.agave'
AGPY_FILENAME = '.agpy'
CACHE_FILENAME = 'current'
SESSIONS_FILENAME = 'config.json'

# Oauth access tokens
TOKEN_SCOPE = 'PRODUCTION'
TOKEN_TTL = 3600



"""AgavePy errors, exceptions, and special handlers
"""

from future import standard_library
standard_library.install_aliases()  # noqa
from urllib.error import HTTPError  # noqa

__all__ = [
    'AgaveError', 'AgaveException', '__handle_tapis_error',
    '_handle_tapis_error'
]


class AgaveError(Exception):
    pass


class AgaveException(Exception):
    pass


def _handle_tapis_error(h):
    if h.code not in (400, 404):
        h.msg = h.msg + ' [{0}]'.format(h.response.text)
    raise h


def __handle_tapis_error(http_error_object):
    """Raise a more detailed HTTPError from Tapis error response
    """
    h = http_error_object
    # extract HTTP response code
    code = -1
    try:
        code = h.response.status_code
        assert isinstance(code, int)
    except Exception:
        # we have no idea what happened
        code = 418

    # extract HTTP reason
    reason = 'UNKNOWN ERROR'
    try:
        reason = h.response.reason
    except Exception:
        pass

    # Tapis APIs will give JSON responses if the target web service is at all
    # capable of fulfilling the request. Therefore, try first to extract fields
    # from the JSON response, then fall back to returning the plain text from
    # the response.
    err_msg = 'Unexpected encountered by the web service'
    status_msg = 'error'
    version_msg = 'unknown'
    try:
        j = h.response.json()
        if 'message' in j:
            err_msg = j['message']
        if 'status' in j:
            status_msg = j['status']
        if 'version' in j:
            version_msg = j['version']
    except Exception:
        err_msg = h.response.text

    httperror = '{0}'.format(err_msg)
    raise HTTPError(code, httperror)



"""Utilities for working with Tapis tenants
"""
from agavepy.settings import TAPIS_TENANTS_URL
import requests

__all__ = ['list_tenants', 'api_server_by_id', 'id_by_api_server']


# TODO - Cache with functools lru_cache
def list_tenants(url=None):
    """List available Tapis tenants

    Returns all tenants configured in a Tapis instance.

    PARAMETERS
    ----------
    url: string (default: settings.TAPIS_TENANTS_URL)
        URL to send GET request to. This resource should list all tenants.

    RETURNS
    -------
    JSON response: dict
        If request was successful, return the json response as a dict.
    """
    if url is None:
        url = TAPIS_TENANTS_URL

    try:
        resp = requests.get(url)
        resp.raise_for_status()
    except Exception:
        raise

    return resp.json().get('result', [])


def api_server_by_id(tenant_id, url=None):
    """Resolve Tapis tenant ID to its API server
    """
    tenants = list_tenants(url)
    for t in tenants:
        if t.get('code').lower() == tenant_id.lower():
            return t.get('baseUrl')


def id_by_api_server(api_server, url=None):
    """Resolve tenant id by its API server
    """
    tenants = list_tenants(url)
    for t in tenants:
        # TODO - ignore presence/absence of trailing slash
        if t.get('baseUrl').lower() == api_server.lower():
            return t.get('code')



#
# Copyright (c) 2013, Digium, Inc.
#
"""Swagger processing libraries.

More information on Swagger can be found `on the Swagger website
<https://developers.helloreverb.com/swagger/>`
"""

__all__ = ["client", "codegen", "processors", "swagger_model"]

from .swagger_model import load_file, load_json, load_url, Loader
from .processors import SwaggerProcessor, SwaggerError
